@@ -7,9 +7,26 @@ mod num; pub(crate) use num::*;
 mod r#enum; pub(crate) use r#enum::*;
 mod seq; pub(crate) use seq::*;
 
+#[derive(Debug,PartialEq)]
+pub(crate) enum Position {
+	Start,
+	Inline,
+	Newline,
+	Indented,
+}
+
+#[derive(Debug,PartialEq)]
+pub(crate) enum ReadResult {
+	Inline,
+	Multiline,
+	None,
+}
+
 pub struct Deserializer<R: buffered_reader::BufferedReader<()>> {
 	input: R,
-	indent: Option<String>,
+	indent: String,
+	prev_indent: Vec<usize>,
+	position: Position,
 	location: Location,
 }
 
@@ -18,165 +35,226 @@ impl<R: buffered_reader::BufferedReader<()>> Deserializer<R> {
 		Self {
 			input: i,
 			indent: Default::default(),
+			prev_indent: Default::default(),
+			position: Position::Start,
 			location: Location{line: 1, column: 1},
 		}
 	}
 
-	fn parse_inline_string(&mut self) -> Result<(std::borrow::Cow<str>, crate::Location), DeserializeError> {
-		let len = self.input.read_to(b'\n')?.len();
-		let mut b = &self.input.consume(len)[..len];
-
-		if let Some(comment_start) = b.iter().position(|c| c == &b'#') {
-			b = &b[..comment_start];
-		}
-
-		let s = crate::from_utf8(b, self.location)?;
-
-		let content = s.trim();
-
-		let mut location = self.location;
-		location.add_columns(content.as_ptr() as usize - s.as_ptr() as usize);
-		self.location.add_lines(1);
-
-		if content.starts_with('"') {
-			return Err(crate::DeserializeErrorKind::Invalid(
-				"Inline strings can't start with `\"`.\n\nQuoted strings are reserved for future features. If you want a string that starts with a quote put the string indented on its own line.".into())
-				.at(location))
-		}
-
-		Ok((std::borrow::Cow::from(content), location))
-	}
-
-	fn parse_indented_string(&mut self) -> Result<(std::borrow::Cow<str>, Location), DeserializeError> {
-		let mut b = self.input.read_to(b'\n')?;
-		let size = b.len();
-		let newline = b.last() == Some(&b'\n');
-		if newline {
-			b = &b[..b.len()-1];
-		}
-		let s = crate::from_utf8(b, self.location)?;
-
-		let content_start = s.find(|c: char| !c.is_whitespace()).unwrap_or(s.len());
-		let (indent, content) = s.split_at(content_start);
-
-		self.location.add_columns(content_start);
-		let location = self.location;
-
-		let root_indent = self.indent.as_ref().map(|s| s.as_str()).unwrap_or("");
-		let prev_root_indent = root_indent.len();
-
-		if !(indent.len() > root_indent.len() && indent.starts_with(root_indent)) {
-			return Ok((std::borrow::Cow::from(""), location))
-		}
-
-		if let Some(ref mut root_indent) = self.indent {
-			root_indent.push_str(&indent[prev_root_indent..]);
-		} else {
-			self.indent = Some(indent.into());
-		}
-
-		let root_indent = self.indent.as_ref().unwrap().as_bytes();
-
-		let mut r = content.to_owned(); // TODO: don't copy for single-line "multiline" block.
-		self.input.consume(size);
-		self.location.add_lines(1);
-
-		loop {
-			let b = self.input.read_to(b'\n')?;
-			let size = b.len();
-
-			if b.is_empty() {
-				break
-			}
-
-			let content = if b.starts_with(root_indent) {
-				let mut content = &b[root_indent.len()..];
-				if content.ends_with(b"\n") {
-					content = &content[..content.len()-1];
-				}
-				crate::from_utf8(content, location)?
-			} else if crate::from_utf8(b, location)?.find(|c: char| !c.is_whitespace()).is_none() {
-				// We don't care about the indent of "empty lines".
-				""
-			} else {
-				break
-			};
-
-			r.push('\n');
-			r.push_str(content);
-
-			self.input.consume(size);
-			self.location.add_lines(1);
-		}
-
-		while r.as_bytes()[r.len() - 1] == b'\n' {
-			r.truncate(r.len() - 1);
-		}
-
-		self.indent.as_mut().unwrap().truncate(prev_root_indent);
-		Ok((std::borrow::Cow::from(r), location))
-	}
-
 	fn parse_string(&mut self) -> Result<(std::borrow::Cow<str>, crate::Location), DeserializeError> {
-		let (s, location) = self.parse_inline_string()?;
-		if !s.is_empty() {
-			// IDK why rust forces us to copy this.
-			Ok((std::borrow::Cow::from(s.into_owned()), location))
-		} else {
-			self.parse_indented_string()
+		let mut r = std::borrow::Cow::Borrowed("");
+		let location = self.location;
+		loop {
+			match self.read_line(false)? {
+				ReadResult::Inline => {
+					return self.parse_inline_string()
+				}
+				ReadResult::Multiline => {
+					let b = self.input.read_to(b'\n')?;
+					let len = b.len();
+
+					let r = r.to_mut();
+					r.push_str(crate::from_utf8(b, self.location)?);
+
+					self.input.consume(len);
+					self.location.add_lines(1);
+				}
+				ReadResult::None => {
+					// TODO: Don't clone.
+					let s = r.to_mut();
+					while s.ends_with('\n') {
+						s.pop().unwrap();
+					}
+
+					self.dedent();
+					return Ok((r, location))
+				}
+			}
 		}
 	}
 
-	fn peek_line(&mut self) -> Result<&str, DeserializeError> {
+	fn parse_inline_string(&mut self) -> Result<(std::borrow::Cow<str>, crate::Location), DeserializeError> {
+		let (s, location, size) = self.peek_inline_string()?;
+		let s = s.to_owned(); // TODO: This is only required so that we can consume the input.
+		self.location.add_lines(1);
+		self.input.consume(size);
+		Ok((s.into(), location))
+	}
+
+	fn peek_inline_string(&mut self) -> Result<(&str, crate::Location, usize), DeserializeError> {
 		let mut b = self.input.read_to(b'\n')?;
-		let mut line_start = b.len();
+		let end = b.len();
 
 		if let Some(comment_start) = b.iter().position(|c| c == &b'#') {
 			b = &b[..comment_start];
 		}
 
-		let s = crate::from_utf8(b, self.location)?;
-		let content = s.trim();
+		let mut location = self.location;
+		let mut s = crate::from_utf8(b, location)?;
 
-		if !content.is_empty() {
-			let content_ptr = content.as_ptr();
-			let content_len = content.len();
-			let d = self.input.buffer();
-			let d = &d[content_ptr as usize - d.as_ptr() as usize..][..content_len];
-			return Ok(unsafe { crate::from_utf8_unchecked(d) })
+		for (i, c) in s.char_indices() {
+			if c.is_whitespace() {
+				location.add_columns(1);
+			} else {
+				s = &s[i..];
+				break
+			}
 		}
 
-		let root_indent = self.indent.as_ref().map(|s| s.as_str()).unwrap_or("");
+		s = s.trim_end();
 
-		let mut offset = line_start;
-		let mut content_start = 0;
-		let mut location = self.location;
+		Ok((s.into(), location, end))
+	}
 
-		loop {
-			b = self.input.data(offset + 1)?;
-			for (i, i_end, c) in bstr::ByteSlice::char_indices(&b[offset..]) {
-				if content_start != 0 && c.is_whitespace() {
-					let d = &self.input.buffer()[content_start..offset+i];
-					return Ok(unsafe { crate::from_utf8_unchecked(d) })
-				} else if c == '\n' {
-					line_start = offset + i_end;
-					location.add_lines(1);
-					continue
-				} else if c == std::char::REPLACEMENT_CHARACTER {
-					crate::from_utf8(&b[i..], self.location)?;
-					unreachable!();
-				} else if content_start == 0 && !c.is_whitespace() {
-					let indent = &b[line_start..offset+i];
-					if !(indent.len() > root_indent.len() && indent.starts_with(root_indent.as_bytes())) {
-						return Ok("")
+	fn parse_key(&mut self) -> Result<(std::borrow::Cow<str>, Location), DeserializeError> {
+		let len = self.input.read_to(b':')?.len();
+		let b = &self.input.consume(len)[..len];
+		let mut s = crate::from_utf8(b, self.location)?;
+
+		if !s.ends_with(':') {
+			return Err(DeserializeErrorKind::Invalid("Expected \":\" after key".into())
+				.at(self.location));
+		}
+
+		s = &s[..s.len() - 1];
+
+		let chars = s.chars().count() + 1;
+		let location = self.location;
+		self.location.add_columns(chars);
+
+		if s.starts_with('\"') {
+			return Err(crate::DeserializeErrorKind::Invalid(
+				"Keys can't start with `\"`.\n\nQuoted strings are reserved for future features. There is currently no way to start a key with `\"`.".into())
+				.at(self.location))
+		}
+
+		self.position = Position::Inline;
+		Ok((s.into(), location))
+	}
+
+	fn read_line(&mut self, comments: bool) -> Result<ReadResult, DeserializeError> {
+		let mut b = self.input.read_to(b'\n')?;
+		let line_end = b.len();
+
+		if b.len() == 0 {
+			// EOF
+			return Ok(ReadResult::None)
+		}
+		if b.ends_with(b"\n") {
+			b = &b[..b.len() - 1];
+		}
+
+		match self.position {
+			Position::Start => {
+				for (i, _i_end, c) in bstr::ByteSlice::char_indices(b) {
+					if c == char::REPLACEMENT_CHARACTER {
+						return Err(DeserializeErrorKind::InvalidUtf8.at(self.location))
 					}
-					content_start = offset + i;
+					if !c.is_whitespace() {
+						let indent = &b[..i];
+						self.indent.push_str(unsafe { crate::from_utf8_unchecked(indent) });
+						self.position = Position::Indented;
+						return self.read_line(comments)
+					}
+				}
+				self.location.add_lines(1);
+				self.input.consume(line_end);
+				self.read_line(comments)
+			}
+			Position::Inline => {
+				for (i, _i_end, c) in bstr::ByteSlice::char_indices(b) {
+					if c == '#' {
+						break
+					} else if c == '"' {
+						return Err(crate::DeserializeErrorKind::Invalid(
+							"Inline strings can't start with `\"`.\n\nQuoted strings are reserved for future features. If you want a string that starts with a quote put the string indented on its own line.".into())
+							.at(self.location))
+					} else if c == char::REPLACEMENT_CHARACTER {
+						self.location.add_columns(i);
+						return Err(DeserializeErrorKind::InvalidUtf8.at(self.location))
+					} else if !c.is_whitespace() {
+						self.input.consume(i);
+						self.position = Position::Indented;
+						return Ok(ReadResult::Inline)
+					}
+
+					self.location.add_columns(1);
 				}
 
-				location.add_columns(1);
+				self.position = Position::Newline;
+				self.input.consume(line_end);
+				self.location.add_lines(1);
+				return self.read_line(comments)
 			}
+			Position::Newline => {
+				let s = crate::from_utf8(b, self.location)?;
+				let indent_len = s.find(|c: char| !c.is_whitespace()).unwrap_or(s.len());
+				let indent = &s[..indent_len];
 
-			offset += b.len();
+				if comments && (s.len() == indent_len || s.as_bytes()[indent_len] == b'#') {
+					self.input.consume(line_end);
+					self.location.add_lines(1);
+					return self.read_line(comments);
+				}
+
+				if indent.len() <= self.indent.len() || !indent.starts_with(&self.indent) {
+					return Ok(ReadResult::None)
+				}
+
+				self.position = Position::Indented;
+				self.prev_indent.push(self.indent.len());
+				self.indent.push_str(&indent[self.indent.len()..]);
+
+				self.location.add_columns(indent.chars().count());
+				self.input.consume(indent_len);
+
+				Ok(ReadResult::Multiline)
+			}
+			Position::Indented => {
+				if comments {
+					let mut empty = true;
+					for (i, _i_end, c) in bstr::ByteSlice::char_indices(b) {
+						if c == '#' {
+							break
+						} else if c == char::REPLACEMENT_CHARACTER {
+							self.location.add_columns(i);
+							return Err(DeserializeErrorKind::InvalidUtf8.at(self.location))
+						} else if !c.is_whitespace() {
+							empty = false;
+							break
+						}
+					}
+
+					if empty {
+						self.input.consume(line_end);
+						self.location.add_lines(1);
+						return self.read_line(comments)
+					}
+				}
+
+				if b.starts_with(self.indent.as_bytes()) {
+					self.input.consume(self.indent.len());
+					self.location.add_columns(self.indent.chars().count());
+					return Ok(ReadResult::Multiline)
+				}
+
+				Ok(ReadResult::None)
+			}
+		}
+	}
+
+	fn dedent(&mut self) {
+		match self.position {
+			Position::Start => unreachable!(),
+			Position::Inline => self.position = Position::Indented,
+			Position::Newline => self.position = Position::Indented,
+			Position::Indented => {
+
+				// Note: Only the top-level object should return None.
+				if let Some(prev) = self.prev_indent.pop() {
+					self.indent.truncate(prev);
+				}
+			}
 		}
 	}
 
@@ -306,21 +384,40 @@ impl<'d, R: buffered_reader::BufferedReader<()>> serde::de::Deserializer<'d> for
 	}
 
 	fn deserialize_option<V: serde::de::Visitor<'d>>(self, visitor: V) -> Result<V::Value, DeserializeError> {
-		match self.peek_line()? {
-			"none" => {
-				// TODO: Fix peek_line to avoid this possibility.
-				let (s, location) = self.parse_string()?;
-				let s = if let Some(comment_start) = s.bytes().position(|c| c == b'#') {
-					&s[..comment_start]
-				} else {
-					&s
-				};
-				if s.trim() != "none" {
-					return Err(crate::DeserializeErrorKind::Unimplemented("Multi line string starting with \"none\" as an option value.").at(location))
-				}
-				visitor.visit_none()
+		if self.position == Position::Inline {
+			let (s, _, size) = self.peek_inline_string()?;
+			if !s.is_empty() {
+				return visitor.visit_some(self)
 			}
-			_ => visitor.visit_some(self), // TODO: Don't consume.
+
+			self.input.consume(size);
+			self.location.add_lines(1);
+			self.position = Position::Newline;
+		}
+
+		let mut b = self.input.read_to(b'\n')?;
+		let size = b.len();
+
+		if size == 0 {
+			return visitor.visit_none()
+		}
+
+		if b.ends_with(b"\n") {
+			b = &b[..b.len()-1];
+		}
+
+		if b.len() > self.indent.len()
+			&& b.starts_with(self.indent.as_bytes())
+			&& bstr::decode_utf8(&b[self.indent.len()..]).0.map(|c| c.is_whitespace()).unwrap_or(false)
+		{
+			visitor.visit_some(self)
+		} else if bstr::ByteSlice::chars(b).all(|c| c.is_whitespace()) {
+			self.input.consume(size);
+			self.location.add_lines(1);
+			self.deserialize_option(visitor)
+		} else {
+			self.dedent();
+			visitor.visit_none()
 		}
 	}
 
@@ -346,12 +443,10 @@ impl<'d, R: buffered_reader::BufferedReader<()>> serde::de::Deserializer<'d> for
 	}
 
 	fn deserialize_seq<V: serde::de::Visitor<'d>>(mut self, visitor: V) -> Result<V::Value, DeserializeError> {
-		let indent_len = self.indent.as_ref().map(|s| s.len()).unwrap_or(0);
 		let r = visitor.visit_seq(SeqDeserializer {
 			root: &mut self,
-			first: true,
 		});
-		self.indent.as_mut().unwrap().truncate(indent_len);
+		self.dedent();
 		r
 	}
 
@@ -369,12 +464,10 @@ impl<'d, R: buffered_reader::BufferedReader<()>> serde::de::Deserializer<'d> for
 	}
 
 	fn deserialize_map<V: serde::de::Visitor<'d>>(mut self, visitor: V) -> Result<V::Value, DeserializeError> {
-		let indent_len = self.indent.as_ref().map(|s| s.len()).unwrap_or(0);
 		let r = visitor.visit_map(MapDeserializer {
 			root: &mut self,
-			first: true,
 		});
-		self.indent.as_mut().unwrap().truncate(indent_len);
+		self.dedent();
 		r
 	}
 
@@ -393,33 +486,40 @@ impl<'d, R: buffered_reader::BufferedReader<()>> serde::de::Deserializer<'d> for
 		variants: &'static [&'static str],
 		visitor: V,
 	) -> Result<V::Value, DeserializeError> {
-		let (inline, location) = self.parse_inline_string()?;
-		if !inline.is_empty() {
-			return serde::de::IntoDeserializer::into_deserializer(inline)
-				.deserialize_enum(name, variants, visitor)
-				.map_err(|e: crate::DeserializeError| e.with_location(location))
-		}
-
-		let s = self.peek_line()?;
-		if s.contains(':') {
-			let indent_len = self.indent.as_ref().map(|s| s.len()).unwrap_or(0);
-			let r = visitor.visit_enum(EnumDeserializer {
-				root: &mut self,
-			});
-			self.indent.as_mut().unwrap().truncate(indent_len);
-			return r
-		}
-
-		let (s, location) = self.parse_indented_string()?;
-		let mut s = s.as_ref();
-		if let Some(comment_start) = s.find('#') {
-			if !s.contains('\n') {
-				s = &s[..comment_start].trim();
+		match self.read_line(true)? {
+			ReadResult::Inline => {
+				let (inline, location) = self.parse_inline_string()?;
+				serde::de::IntoDeserializer::into_deserializer(inline)
+					.deserialize_enum(name, variants, visitor)
+					.map_err(|e: crate::DeserializeError| e.with_location(location))
+			}
+			ReadResult::None => {
+				self.dedent();
+				serde::de::IntoDeserializer::into_deserializer("")
+					.deserialize_enum(name, variants, visitor)
+					.map_err(|e: crate::DeserializeError| e.with_location(self.location))
+			}
+			ReadResult::Multiline => {
+				let mut b = self.input.read_to(b'\n')?;
+				if let Some(comment_start) = b.iter().position(|c| c == &b'#') {
+					b = &b[..comment_start];
+				}
+				if b.contains(&b':') {
+					let r = visitor.visit_enum(EnumDeserializer {
+						root: &mut self,
+					});
+					self.dedent();
+					r
+				} else {
+					let (inline, location) = self.parse_inline_string()?;
+					let r = serde::de::IntoDeserializer::into_deserializer(inline)
+						.deserialize_enum(name, variants, visitor)
+						.map_err(|e: crate::DeserializeError| e.with_location(location));
+					self.dedent();
+					r
+				}
 			}
 		}
-		serde::de::IntoDeserializer::into_deserializer(s)
-			.deserialize_enum(name, variants, visitor)
-			.map_err(|e: crate::DeserializeError| e.with_location(location))
 	}
 
 	fn deserialize_identifier<V: serde::de::Visitor<'d>>(self, visitor: V) -> Result<V::Value, DeserializeError> {
